//
//  Spu eXtended Math library -- DNE
//
#ifndef SXM_H
#define SXM_H

#include <spu_intrinsics.h>

vector unsigned int SXM_ZERO_VEC = { 0, 0, 0, 0 };
vector unsigned int SXM_ONES_VEC = { 1, 1, 1, 1 };


//
// core functions
//

// unsigned multiplication: c = a * b  (c 64 bit; a, b 32 bit)
void sxm_umul( vector unsigned int * ch, vector unsigned int * cl, // output high & low word
               vector unsigned int a,                              // input word
               vector unsigned int b );                            // input word

// unsigned division/modulo: a = c / b resp. a = c % b  (c 64 bit; a, b 32 bit)
vector unsigned int
  sxm_udivmod( vector unsigned int ch, vector unsigned int cl, // input high & low word
               vector unsigned int b,                          // input word (divisor)
               vector unsigned int op );                       // input word op: 0=div, 1=mod,
                                                               //      for each vector element

// unsigned modular exponentiation: c = ( a ^ b ) % m  ( a, b, c, m 32 bit)
vector unsigned int
  sxm_umodpow( vector unsigned int a,                          // input word (base)
               vector unsigned int b,                          // input word (exponent)
               vector unsigned int m );	                       // input word (module)

// unsigned exponentiation: c = a ^ b  ( a, b, c 32 bit)
vector unsigned int
     sxm_upow( vector unsigned int a,                          // input word (base)
               vector unsigned int b );                        // input word (exponent)

// float reciprocal:  y = 1/x  (x, y float)
vector float
   sxm_frecp( vector float x );	                               // imput float

// float division:  z = y/x  (x, y, z float)
vector float
    sxm_fdiv( vector float y, 	                               // imput float
              vector float x );	                               // imput float


//
//  derived functions
//

// unsigned division: a = c / b (c 64 bit; a, b 32 bit)
vector unsigned int
     sxm_udiv( vector unsigned int ch, vector unsigned int cl, // input high & low word
               vector unsigned int b )                         // input word (divisor)
{ return sxm_udivmod( ch, cl, b, SXM_ZERO_VEC ); }

// unsigned modulo: a = c % b (c 64 bit; a, b 32 bit)
vector unsigned int
     sxm_umod( vector unsigned int ch, vector unsigned int cl, // input high & low word
               vector unsigned int b )                         // input word (divisor)
{ return sxm_udivmod( ch, cl, b, SXM_ONES_VEC ); }

#endif
