#
#  Spu eXtended Math library -- Diane Neisius (DNE) 3/2010
#


# float 32 bit reciprocal with 32 bit result
# -- atomic function, needs no special care for stack --
#
#  R3: input float vector
#
#  The algorithm bases on the Newton reciprocal algorithm,
#  iteration loop (5 times for 32 bit precision) unrolled

	.data
	.align 4
fp1:	.word	0x3f800000, 0x3f800000, 0x3f800000, 0x3f800000	# float 1.0

	.text
	.align 4
	.global sxm_frecp
	.type	sxm_frecp, @function

sxm_frecp:				# P  comment
	or	$4,  $3,  $3		# 0  move R3, R4
	frest	$3,  $4			# 0
	nop				# 0
	lqa	$22, fp1		# 1  load 1.0

	fnms	$20, $3,  $4,  $22	# 0  R20 = -(a*x - 1.0) is the 1st error
	fma	$3,  $3,  $20, $3	# 0  R3  = x*R20 + x
	fnms	$20, $3,  $4,  $22	# 0  R20 = -(a*x - 1.0) is the 2nd error
	fma	$3,  $3,  $20, $3	# 0  R3  = x*R20 + x

	fnms	$20, $3,  $4,  $22	# 0  R20 = -(a*x - 1.0) is the 3rd error
	fma	$3,  $3,  $20, $3	# 0  R3  = x*R20 + x
	fnms	$20, $3,  $4,  $22	# 0  R20 = -(a*x - 1.0) is the 4th error
	fma	$3,  $3,  $20, $3	# 0  R3  = x*R20 + x

	fnms	$20, $3,  $4,  $22	# 0  R20 = -(a*x - 1.0) is the 5th error
	bi	$lr   			# 1


# float 32 bit division with 32 bit result
# -- non-atomic function, needs special care for stack --
#
#  R3: input float vector
#  R4: input float vector
#
#  uses sxm_frecp above
#
#  local stack frame:
#
#	0($sp)	old stack pointer
#	16($sp)	return adress
#	32($sp) R3 saved
#

	.text
	.align 4
	.global sxm_fdiv
	.type	sxm_fdiv, @function

sxm_fdiv:				# P  comment
	nop				# 0
	stqd	$sp, -48($sp)		# 1  care for stack
	ai	$sp, $sp, -48		# 0
	stqd	$lr, 16($sp)		# 1

	nop				# 0
	stqd	$3,  32($sp)		# 1
	or	$3,  $4, $4		# 0
	brsl	$lr, sxm_frecp		# 1

	nop				# 0
	lqd	$4,  32($sp)		# 1
	fm	$3,  $3,  $4		# 0
	lqd	$lr, 16($sp)		# 1  restore return address

	ai	$sp, $sp, 48		# 0  cleanup stack
	bi	$lr			# 1  return

