#
#  Spu eXtended Math library -- Diane Neisius (DNE) 3/2010
#


# unsigned exponentiation 32 by 32 bit with 32 bit result
# -- non-atomic function, needs care for stack --
#
#  R3: input base word vector / result vector at exit
#  R4: input exponent word vector
#
#  The algorithm is a simplified version of modpow (see below)


	.text
	.align 4
	.global sxm_upow
	.type   sxm_upow, @function

mpy32:					# P  comment
	mpyh	$20, $3,  $4		# 0  the 32 bit unsigned multiply routine
	mpyh	$21, $4,  $3		# 0  from the IBM spu assembler manual :)
	mpyu	$22, $3,  $4		# 0
	a	$3,  $20, $21		# 0
	a	$3,  $3,  $22		# 0
	bi	$lr			# 1

sxm_upow:				# P  comment:
	il	$71, 1			# 0
	stqd	$sp, -112($sp)		# 1  new stack frame
	ai	$sp, $sp, -112		# 0 
	stqd	$lr, 16($sp)		# 1  save return address
 
	nop				# 0
	stqd	$3,  32($sp)		# 1  save base (par.)
	stqd	$4,  48($sp)		# 1  save exp  (par.)
	stqd	$71, 64($sp)		# 1  init res = 1    (loc.)
	
loopp:
	nop				# 0
	lqd	$30, 48($sp)		# 1  load exp
	ceqi	$31, $30, 0		# 0  exp == 0 ?
	lqd	$3,  64($sp)		# 1  load res

	nor	$31, $31, $31		# 0
	stqd	$31, 80($sp)		# 1  store while_mask ( while (exp) {...} )
	andi	$32, $30, 1		# 0
	lqd	$4,  32($sp)		# 1  load base

	ceqi	$32, $32, 1		# 0  exp & 1
	stqd	$32, 96($sp)		# 1  store if_mask  ( if ( exp & 1 ) {...} )
	nop				# 0
	brsl	$lr, mpy32		# 1  R3 = base * res

	nop				# 0
	lqd	$30, 80($sp)		# 1  while ...
	nop				# 0
	lqd	$31, 96($sp)		# 1  if ...

	and	$30, $30, $31		# 0
	lqd	$32, 64($sp)		# 1  load res
	and	$33, $3,  $30		# 0
	lqd	$3,  32($sp)		# 1  load base

	andc	$32, $32, $30		# 0
	lnop				# 1
	or	$33, $33, $32		# 0  new res
	stqd	$33, 64($sp)		# 1  store res

	lqd	$4,  32($sp)		# 1  load base
	brsl	$lr, mpy32		# 1  R3 = base * base
	lqd	$30, 80($sp)		# 1  while ...
	lqd	$32, 32($sp)		# 1  load base

	and	$3,  $3,  $30		# 0
	lqd	$35, 80($sp)		# 1  load while_mask
	andc	$34, $34, $30		# 0
	lqd	$31, 48($sp)		# 1  load exp

	or	$3,  $3,  $34		# 0  new base
	stqd	$3,  32($sp)		# 1  store base
	rotmi	$32, $31, -1		# 0  exp >>= 1
	orx	$34, $35		# 1  all while loops finished?

	and	$32, $32, $35		# 0
	andc	$31, $31, $35		# 0
	or	$33, $31, $32		# 0
	stqd	$33, 48($sp)		# 1  store exp

	nop				# 0
	brnz	$34, loopp		# 1
	nop				# 0  18 cycles pipeline stall
	lqd	$3, 64($sp)		# 1  return res
	
	nop				# 0
	lqd	$lr, 16($sp)		# 1  restore return address
	ai	$sp, $sp, 112		# 0  cleanup stack
	bi	$lr			# 1  return


# unsigned modular exponentiation 32 by 32 bit with 32 bit result
# -- non-atomic function, needs care for stack --
#
#  R3: input base word vector / result vector at exit
#  R4: input exponent word vector
#  R5: input module word vector
#
#  The algorithm is inspired by 'modpow' of Wikipedia (see "modular exponentiation")

	.data
	.align 4
tmp64h:	.word	0, 0, 0, 0		# temporary store for intermediate 64 bit unsigned
tmp64l:	.word	0, 0, 0, 0

	.text
	.align 4
	.global sxm_umodpow
	.type   sxm_umodpow, @function

sxm_umodpow:				# P  comment:
	il	$71, 1			# 0
	stqd	$sp, -128($sp)		# 1  new stack frame
	ai	$sp, $sp, -128		# 0 
	stqd	$lr, 16($sp)		# 1  save return address
 
	stqd	$3,  32($sp)		# 1  save base (par.)
	stqd	$4,  48($sp)		# 1  save exp  (par.)
	stqd	$5,  64($sp)		# 1  save mod  (par.)
	stqd	$71, 112($sp)		# 1  init res = 1    (loc.)
	
loopm:
	ila	$3,  tmp64h		# 0  pointer to temp 64 bit store
	lqd	$30, 48($sp)		# 1  load exp
	ceqi	$31, $30, 0		# 0  exp == 0 ?
	lqd	$5,  112($sp)		# 1  load res

	nor	$31, $31, $31		# 0
	stqd	$31, 80($sp)		# 1  store while_mask ( while (exp) {...} )
	andi	$32, $30, 1		# 0
	lqd	$6,  32($sp)		# 1  load base

	ceqi	$32, $32, 1		# 0  exp & 1
	stqd	$32, 96($sp)		# 1  store if_mask  ( if ( exp & 1 ) {...} )
	ila	$4,  tmp64l		# 0  pointer to temp 64 bit store
	brsl	$lr, sxm_umul		# 1  tmp64 = base * res

	il	$6,  1			# 0  op: modulo
	lqa	$3,  tmp64h		# 1  load tmp64
	nop				# 0
	lqa	$4,  tmp64l		# 1  load tmp64

	lqd	$5,  64($sp)		# 1  load mod
	brsl	$lr, sxm_udivmod	# 1  R3 = ( base * res ) % mod
	lqd	$30, 80($sp)		# 1  while ...
	lqd	$31, 96($sp)		# 1  if ...

	and	$30, $30, $31		# 0
	lqd	$32, 112($sp)		# 1  load res
	and	$33, $3,  $30		# 0
	lqd	$5,  32($sp)		# 1  load base

	andc	$32, $32, $30		# 0
	lnop				# 1
	or	$33, $33, $32		# 0  new res
	stqd	$33, 112($sp)		# 1  store res

	ila	$3,  tmp64h		# 0  pointer to temp 64 bit store
	lqd	$6,  32($sp)		# 1  load base
	ila	$4,  tmp64l		# 0  pointer to temp 64 bit store
	brsl	$lr, sxm_umul		# 1  tmp64 = base * base

	il	$6,  1			# 0  op: modulo
	lqa	$3,  tmp64h		# 1  load tmp64
	nop				# 0
	lqa	$4,  tmp64l		# 1  load tmp64


	lqd	$5,  64($sp)		# 1  load mod
	brsl	$lr, sxm_udivmod	# 1  R3 = ( base * base ) % mod
	lqd	$30, 80($sp)		# 1  while ...
	lqd	$32, 32($sp)		# 1  load base

	and	$3,  $3,  $30		# 0
	lqd	$35, 80($sp)		# 1  load while_mask
	andc	$34, $34, $30		# 0
	lqd	$31, 48($sp)		# 1  load exp

	or	$3,  $3,  $34		# 0  new base
	stqd	$3,  32($sp)		# 1  store base
	rotmi	$32, $31, -1		# 0  exp >>= 1
	orx	$34, $35		# 1  all while loops finished?

	and	$32, $32, $35		# 0
	andc	$31, $31, $35		# 0
	or	$33, $31, $32		# 0
	stqd	$33, 48($sp)		# 1  store exp

	nop				# 0
	brnz	$34, loopm		# 1
	nop				# 0  18 cycles pipeline stall
	lqd	$3, 112($sp)		# 1  return res
	
	nop				# 0
	lqd	$lr, 16($sp)		# 1  restore return address
	ai	$sp, $sp, 128		# 0  cleanup stack
	bi	$lr			# 1  return
